/*-*-C-*-
 * Copyright 2006  Petter Urkedal
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef CUEX_BINDING_H
#define CUEX_BINDING_H

#include <cuex/fwd.h>
#include <cuex/var.h>

CU_BEGIN_DECLARATIONS
/*!\defgroup cuex_binding_h cuex/binding.h: Variable Binding using de Bruijn Indices
 *@{\ingroup cuex_mod
 *
 * This provides variables, scoping constructs, and algorithms for expressions
 * using de Bruijn indexes.  This includes the built-in μ-recursion and
 * λ-expressions.
 *
 * \see cuex_recursion_h
 */

#define cuex_bvarmeta(index) \
    cuex_varmeta_kqi(cuex_varkind_bvar, cuex_qcode_n, index)

/*!True iff \a meta is that of a b-variable. */
CU_SINLINE cu_bool_t cuex_is_bvarmeta(cuex_meta_t meta)
{ return cuex_is_varmeta_k(meta, cuex_varkind_bvar); }

/*!The de Bruijn index of \a bvarmeta.
 * \pre \a bvarmeta is the meta of a b-variable. */
CU_SINLINE unsigned int cuex_bvarmeta_index(cuex_meta_t bvarmeta)
{ return cuex_varmeta_index(bvarmeta); }

/*!Returns a b-variable with de Bruijn index \a index. */
CU_SINLINE cuex_t cuex_bvar(unsigned int index)
{ return cuex_halloc_by_key(cuex_bvarmeta(index), 0, NULL); }

cuex_t cuexP_bfree_adjusted(cuex_t e, int l_diff, int l_top);

/*!For all free b-variables in \a e, increment their de Bruijn index by \a
 * l_diff.  This prepares the subexpression \a e to be substituted down
 * accross \a l_diff bind-sites, or for negative, up accross -\a l_diff
 * bind-sites. */
CU_SINLINE cuex_t cuex_bfree_adjusted(cuex_t e, int l_diff)
{ return cuexP_bfree_adjusted(e, l_diff, -1); }


/*!@}*/
CU_END_DECLARATIONS

#endif
